#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "1.0"
__author__    = "ed <filedrop@ocv.me>"
__url__       = "https://ocv.me/dev/?filedrop.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019

# for a smaller (and much slower) alternative,
#   see https://ocv.me/dev/?filesink.py
#
# expect 305 MB/s with python3, or 356 MB/s using python2,
# that's like 11x faster than the built-in http parser at 32 MB/s
# (which is what filesink.py uses, cgi.FieldStorage)
#
# navigate to your ip:port combination with a browser to see a
# basic upload form, or just use curl to upload windows7.iso:
#   curl -F "f=@windows7.iso" http://127.0.0.1:1234/
#
# the http response is a minimal html which contains the
# filesize and sha1 checksum of the uploaded file, as seen by us


import os
import sys
import time
import hashlib
import select
import socket
import logging

logging.basicConfig(
	level = logging.DEBUG,
	format = '\033[36m%(asctime)s \033[33m%(levelname)5s \033[0m%(message)s'
)

if sys.version_info[0] > 2:
	unicode = str


class UnbufferedHttpd(object):
	def __init__(self, addr, port):
		srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		srv.bind((addr, port))
		srv.listen(3)
		
		while True:
			UnbufferedClient(*srv.accept()).run()


class UnbufferedClient(object):
	def __init__(self, sck, addr):
		self.s = sck
		self.addr = addr
		self.ok = True
	
	def run(self):
		headerlines = self.read_header()
		if not self.ok:
			return
		
		self.headers = {}
		mode, self.req, _ = headerlines[0].split(' ')
		
		for header_line in headerlines[1:]:
			k, v = header_line.split(':', 1)
			self.headers[k.lower()] = v.strip()
		
		if mode == 'GET':
			self.handle_get()
		elif mode == 'POST':
			self.handle_post()
		else:
			self.loud_reply('invalid HTTP mode "{0}"'.format(mode))
	
	def panic(self, msg):
		logging.error('client disconnected ({0})'.format(msg).upper())
		self.ok = False
		self.s.close()
	
	def read_header(self):
		ret = b''
		while True:
			if ret.endswith(b'\r\n\r\n'):
				break
			elif ret.endswith(b'\r\n\r'):
				n = 1
			elif ret.endswith(b'\r\n'):
				n = 2
			elif ret.endswith(b'\r'):
				n = 3
			else:
				n = 4
			
			buf = self.s.recv(n)
			if not buf:
				self.panic('headers')
				break
			
			ret += buf
		
		return ret[:-4].decode('utf-8', 'replace').split('\r\n')
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\n\r\n'.format(len(body)).encode('utf-8')
		if self.ok:
			self.s.send(header + body)
		
		self.s.close()
		return body
	
	def loud_reply(self, body):
		logging.info(body.decode('utf-8').rstrip())
		self.reply(body)
	
	def handle_get(self):
		logging.info('')
		logging.info('GET  {0} {1}'.format(self.addr[0], self.req))
		self.reply(b'<form method="post" enctype="multipart/form-data"><h5><input type="file" name="f"><h5><input type="submit" value="start upload">')
	
	def handle_post(self):
		logging.info('')
		logging.info('POST {0} {1}'.format(self.addr[0], self.req))
		
		try:
			if self.headers['expect'].lower() == '100-continue':
				self.s.send(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		form_segm = self.read_header()
		if not self.ok:
			return
		
		boundary = b'\r\n' + form_segm[0].encode('utf-8') + b'--\r\n'
		for ln in form_segm[1:]:
			logging.debug(ln)
		
		fn = 'inc.{0:.3f}'.format(time.time())
		logging.info('writing to {0}'.format(fn))
		#fn = '/dev/null'

		t0 = time.time()
		with open(fn, 'wb') as f:
			sz, sha1 = self.handle_multipart(boundary, f)
		
		td = time.time() - t0
		spd = (sz / td) / (1024*1024)
		
		status = 'OK'
		if not self.ok:
			status = 'ERROR'
		
		msg = '{0} // {1} bytes // {2:.3f} MiB/s<br> sha1: {3}\n'.format(status, sz, spd, sha1)
		self.loud_reply(msg.encode('utf-8'))

		if fn != '/dev/null':
			with open(fn + '.txt', 'wb') as f:
				f.write((u'\n'.join(unicode(x) for x in [
					u':'.join(unicode(x) for x in self.addr),
					u'\n'.join(form_segm),
					msg.rstrip()
					]) + '\n').encode('utf-8'))
	
	def handle_multipart(self, boundary, ofd):
		tlen = 0
		hashobj = hashlib.sha1()
		for buf in self.extract_filedata(boundary):
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def extract_filedata(self, boundary):
		blen = len(boundary)
		while True:
			buf = self.s.recv(4096)
			if not buf:
				# abort: client disconnected
				self.panic('outer')
				return
			
			while True:
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in boundary:
						n -= 1
						break
				
				if n == 0 \
				or not boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.s.recv(4096)
				if not buf2:
					# abort: client disconnected
					self.panic('inner')
					return
				
				buf += buf2
			
			yield buf


port = 1234
if len(sys.argv) > 1:
	port = int(sys.argv[1])

logging.info('listening @ http://127.0.0.1:{0}/'.format(port))
UnbufferedHttpd('0.0.0.0', port)
